package ij.process;

import java.awt.*;
import java.awt.image.*;
import ij.*;
import ij.gui.*;
import ij.measure.*;

/** This class does image type conversions. */
public class ImageConverter {
	private ImagePlus imp;
	private int type;
	private static boolean doScaling = Prefs.getBoolean(Prefs.SCALE_CONVERSIONS,true);

	public ImageConverter(ImagePlus imp) {
		this.imp = imp;
		type = imp.getType();
	}

	/** Converts this ImagePlus to 8-bit grayscale. */
	public synchronized void convertToGray8() {
		if (imp.getStackSize()>1)
			throw new IllegalArgumentException("Unsupported conversion");

		if (type==ImagePlus.GRAY16 || type==ImagePlus.GRAY32) {
			ImageProcessor ip = imp.getProcessor();
			imp.setProcessor(null, convertToByteProcessor(ip));
			imp.getCalibration().disableDensityCalibration();
		} else {
			ImageProcessor ip;
			if (imp.getType()==ImagePlus.COLOR_RGB)
				ip = imp.getProcessor();
			else
				ip = new ColorProcessor(imp.getImage());
	    	imp.setProcessor(null, convertToByteProcessor(ip));
	    }
	}

	/** Converts a ColorProcessor, ShortProcessor or FloatProcessor to a ByteProcessor. */
	public static ByteProcessor convertToByteProcessor(ImageProcessor ip) {
		if (ip instanceof ShortProcessor)
			return convertShortProcessorToByteProcessor((ShortProcessor)ip);
		else if (ip instanceof FloatProcessor)
			return convertFloatProcessorToByteProcessor((FloatProcessor)ip);
		else if (ip instanceof ColorProcessor)
			return convertColorProcessorToByteProcessor((ColorProcessor)ip);
		else return null;
	}

	/** Converts a ShortProcessor to a ByteProcessor. */
	public static ByteProcessor convertShortProcessorToByteProcessor(ShortProcessor ip) {
		if (doScaling) {
			Image img = ip.createImage();
			return new ByteProcessor(img);
		} else {
			int width = ip.getWidth();
			int height = ip.getHeight();
			short[] pixels16 = (short[])ip.getPixels();
			byte[] pixels8 = new byte[width*height];
			int value;
			for (int i=0; i<width*height; i++) {
				value = pixels16[i]&0xffff;
				if (value>255) value = 255;
				pixels8[i] = (byte)value;
			}
			return new ByteProcessor(width, height, pixels8, ip.getColorModel());
		}
	}

	/** Converts a FloatProcessor to a ByteProcessor. */
	public static ByteProcessor convertFloatProcessorToByteProcessor(FloatProcessor ip) {
		if (doScaling) {
			Image img = ip.createImage();
			return new ByteProcessor(img);
		} else {
			int width = ip.getWidth();
			int height = ip.getHeight();
			float[] pixels32 = (float[])ip.getPixels();
			byte[] pixels8 = new byte[width*height];
			float value;
			for (int i=0; i<width*height; i++) {
				value = pixels32[i];
				if (value<0f) value = 0f;
				if (value>65535f) value = 65535f;
				pixels8[i] = (byte)Math.round(value);
			}
			return new ByteProcessor(width, height, pixels8, ip.getColorModel());
		}
	}

	/** Converts a ColorProcessor to a ByteProcessor. */
	public static ByteProcessor convertColorProcessorToByteProcessor(ColorProcessor ip) {
		int c, r, g, b;
		int width, height;
		int[] pixels32;
		byte[] pixels8;
		ColorModel cm;
		Image img8;
		
		//get RGB pixels
		width = ip.getWidth();
		height = ip.getHeight();
		pixels32 = (int[])ip.getPixels();
		
		//convert to grayscale
		pixels8 = new byte[width * height];
		for (int i=0; i < width*height; i++) {
			c = pixels32[i];
			r = (c&0xff0000)>>16;
			g = (c&0xff00)>>8;
			b = c&0xff;
			pixels8[i] = (byte)((int)(r*0.30 + g*0.59 + b*0.11) & 0xff);
		}
		
		return new ByteProcessor(width, height, pixels8, null);
	}

	/** Converts this ImagePlus to 16-bit grayscale. */
	public void convertToGray16() {
		if (type==ImagePlus.GRAY32)
			{convertGray32ToGray16(); return;};
		if (type!=ImagePlus.GRAY8)
			throw new IllegalArgumentException("Unsupported conversion");
		int width = imp.getWidth();
		int height = imp.getHeight();
		ImageProcessor ip = imp.getProcessor();
		byte[] pixels8 = (byte[])ip.getPixels();
		boolean invertedLut = imp.isInvertedLut();
		imp.trimProcessor();
		short[] pixels16 = new short[width * height];
		int inc = (width*height)/20;
		if (inc<1) inc = 1;
		for (int i=0,j=0; i<width*height; i++) {
			pixels16[i] = (short)(pixels8[i]&0xff);
			if (j++==inc)
				{IJ.showProgress(i/(double)(width*height)); j=0;}
		}
	    ImageProcessor ip16 = new ShortProcessor(width, height, pixels16, ip.getColorModel());
		imp.setProcessor(null, ip16);
		imp.getCalibration().disableDensityCalibration();
	}

	/** Converts this ImagePlus from 32-bit float to 16-bit 
		signed grayscale, scaling to the full 16-bit range. */
	public void convertGray32ToGray16() {
		if (type!=ImagePlus.GRAY32)
			throw new IllegalArgumentException("Unsupported conversion");
		int width = imp.getWidth();
		int height = imp.getHeight();
		ImageProcessor ip = imp.getProcessor();
		float[] pixels32 = (float[])ip.getPixels();
		boolean invertedLut = imp.isInvertedLut();
		short[] pixels16 = new short[width*height];
		int inc = (width*height)/20;
		if (inc<1) inc = 1;
		double min = ip.getMin();
		double max = ip.getMax();
		double scale;
		if ((max-min)==0.0)
			scale = 1.0;
		else
			scale = 65535.0/(max-min);
		double value;
		for (int i=0,j=0; i<width*height; i++) {
			if (doScaling)
				value = (pixels32[i]-min)*scale;
			else
				value = pixels32[i];
			if (value<0.0) value = 0.0;
			if (value>65535.0) value = 65535.0;
			pixels16[i] = (short)value;
			if (j++==inc)
				{IJ.showProgress(i/(double)(width*height)); j=0;}
		}
	    ColorModel cm = ip.getColorModel();
	    ImageProcessor ip16 = new ShortProcessor(width, height, pixels16, cm);
		imp.setProcessor(null, ip16);
		Undo.reset(); // can't undo
	}

	/** Converts this ImagePlus to 32-bit (float) grayscale. */
	public void convertToGray32() {
		if (type==ImagePlus.GRAY16)
			{convertGray16ToGray32(); return;}
		if (type!=ImagePlus.GRAY8)
			throw new IllegalArgumentException("Unsupported conversion");
		ByteProcessor bp = (ByteProcessor)imp.getProcessor();
		Calibration cal = imp.getCalibration();
	    FloatProcessor fp = convertByteToFloat(bp, cal);
		imp.setProcessor(null, fp);
		cal.disableDensityCalibration();
	}

	/** Converts this ImagePlus from 16-bit grayscale
		to 32-bit (float) grayscale. */
	public void convertGray16ToGray32() {
		if (type!=ImagePlus.GRAY16)
			throw new IllegalArgumentException("Unsupported conversion");
		ShortProcessor sp = (ShortProcessor)imp.getProcessor();
		Calibration cal = imp.getCalibration();
	    FloatProcessor fp = convertShortToFloat(sp, cal);
		imp.setProcessor(null, fp);
		cal.disableDensityCalibration();
		Undo.reset(); // can't undo
	}
	
	/** Converts a ByteProcessor to a FloatProcessor.  Uses a calibration
		table if cal!=null and it contains a calibration table. */
	public static FloatProcessor convertByteToFloat(ByteProcessor bp, Calibration cal) {
		int width = bp.getWidth();
		int height = bp.getHeight();
		byte[] pixels8 = (byte[])bp.getPixels();
		boolean invertedLut = bp.isInvertedLut();
		float[] pixels32 = new float[width*height];
		int value;
		float[] cTable = null;
		if (cal!=null)
			cTable = cal.getCTable();
		if (cTable!=null && cTable.length==256)
			for (int i=0; i<width*height; i++)
				pixels32[i] = cTable[pixels8[i]&255];
		else
			for (int i=0; i<width*height; i++)
				pixels32[i] = pixels8[i]&255;
	    ColorModel cm = bp.getColorModel();
	    return new FloatProcessor(width, height, pixels32, cm);
	}

	/** Converts a ShortProcessor to a FloatProcessor. Uses a calibration
		table if cal!=null and it contains a calibration table. */
	public static FloatProcessor convertShortToFloat(ShortProcessor sp, Calibration cal) {
		int width = sp.getWidth();
		int height = sp.getHeight();
		short[] pixels16 = (short[])sp.getPixels();
		boolean invertedLut = false; //imp.isInvertedLut();
		float[] pixels32 = new float[width*height];
		int value;
		float[] cTable = null;
		if (cal!=null)
			cTable = cal.getCTable();
		if (cTable!=null && cTable.length==65536)
			for (int i=0; i<width*height; i++)
				pixels32[i] = cTable[pixels16[i]&0xffff];
		else
			for (int i=0; i<width*height; i++)
				pixels32[i] = pixels16[i]&0xffff;
	    ColorModel cm = sp.getColorModel();
	    return new FloatProcessor(width, height, pixels32, cm);
	}

	/** Converts an RGB image to an RGB (red, green and blue) stack. */
	public void convertToRGBStack() {
		if (type!=ImagePlus.COLOR_RGB)
			throw new IllegalArgumentException("Image must be RGB");

		//convert to RGB Stack
		ColorProcessor cp;
		if (imp.getType()==ImagePlus.COLOR_RGB)
			cp = (ColorProcessor)imp.getProcessor();
		else
			cp = new ColorProcessor(imp.getImage());
		int width = imp.getWidth();
		int height = imp.getHeight();
		byte[] R = new byte[width*height];
		byte[] G = new byte[width*height];
		byte[] B = new byte[width*height];
		cp.getRGB(R, G, B);
		imp.trimProcessor();
		
		// Create stack and select Red channel
		ColorModel cm = LookUpTable.createGrayscaleColorModel(false);
		ImageStack stack = new ImageStack(width, height, cm);
		stack.addSlice("Red", R);
		stack.addSlice("Green", G);
		stack.addSlice("Blue", B);
		imp.setStack(null, stack);
	}

	/** Converts an RGB image to a HSB (hue, saturation and brightness) stack. */
	public void convertToHSB() {
		if (type!=ImagePlus.COLOR_RGB)
			throw new IllegalArgumentException("Image must be RGB");

		//convert to hue, saturation and brightness
		//IJ.showProgress(0.1);
		ColorProcessor cp;
		if (imp.getType()==ImagePlus.COLOR_RGB)
			cp = (ColorProcessor)imp.getProcessor();
		else
			cp = new ColorProcessor(imp.getImage());
		int width = imp.getWidth();
		int height = imp.getHeight();
		byte[] H = new byte[width*height];
		byte[] S = new byte[width*height];
		byte[] B = new byte[width*height];
		cp.getHSB(H, S, B);
		imp.trimProcessor();

		//create stack and select hue channel
		ColorModel cm = LookUpTable.createGrayscaleColorModel(false);
		ImageStack stack = new ImageStack(width, height, cm);
		stack.addSlice("Hue", H);
		stack.addSlice("Saturation", S);
		stack.addSlice("Brightness", B);
		imp.setStack(null, stack);
		//IJ.showProgress(1.0);
	}
	
	/** Converts image to RGB. */
	public void convertToRGB() {
		if (type==ImagePlus.GRAY16 || type==ImagePlus.GRAY32)
			convertToGray8();
		type = imp.getType();
		if (type==ImagePlus.GRAY8 || type==ImagePlus.COLOR_256) 
			convertEightBitToRGB();
		else	
			throw new IllegalArgumentException("Unsupported conversion");
	}
	
	/** Converts an 8-bit image to RGB. */
	public void convertEightBitToRGB() {
		if (!(type==ImagePlus.GRAY8 || type==ImagePlus.COLOR_256))
			throw new IllegalArgumentException("Image must be 8-bits");
		int width = imp.getWidth();
		int height = imp.getHeight();
		ColorProcessor cp = new ColorProcessor(imp.getImage());
		imp.setImage(cp.createImage());
		imp.getCalibration().disableDensityCalibration();
	}
	
	/** Converts a 3 slice 8-bit stack to RGB. */
	public void convertRGBStackToRGB() {
		int stackSize = imp.getStackSize();
		if (stackSize<3 ||stackSize>4 || type!=ImagePlus.GRAY8)
			throw new IllegalArgumentException("3-slice 8-bit stack required");
		int width = imp.getWidth();
		int height = imp.getHeight();
		ImageStack stack = imp.getStack();
		byte[] R = (byte[])stack.getPixels(1);
		byte[] G = (byte[])stack.getPixels(2);
		byte[] B = (byte[])stack.getPixels(3);
		imp.trimProcessor();
		ColorProcessor cp = new ColorProcessor(width, height);
		cp.setRGB(R, G, B);
		if (imp.isInvertedLut())
			cp.invert();
		imp.setImage(cp.createImage());
		imp.killStack();
	}

	/** Converts a 3-slice (hue, saturation, brightness) 8-bit stack to RGB. */
	public void convertHSBToRGB() {
		if (imp.getStackSize()!=3)
			throw new IllegalArgumentException("3-slice 8-bit stack required");
		ImageStack stack = imp.getStack();
		byte[] H = (byte[])stack.getPixels(1);
		byte[] S = (byte[])stack.getPixels(2);
		byte[] B = (byte[])stack.getPixels(3);
		int width = imp.getWidth();
		int height = imp.getHeight();
		imp.trimProcessor();
		ColorProcessor cp = new ColorProcessor(width, height);
		cp.setHSB(H, S, B);
		imp.setImage(cp.createImage());
		imp.killStack();
	}
	
	/** Converts an RGB image to 8-bits indexed color. 'nColors must
		be greater than 1 and less than or equal to 256. */
	public void convertRGBtoIndexedColor(int nColors) {
		if (type!=ImagePlus.COLOR_RGB)
			throw new IllegalArgumentException("Image must be RGB");
		if (nColors<2) nColors = 2;
		if (nColors>256) nColors = 256;
		
		// get RGB pixels
		IJ.showProgress(0.1);
		IJ.showStatus("Grabbing pixels");
		int width = imp.getWidth();
		int height = imp.getHeight();
		ImageProcessor ip = imp.getProcessor();
	 	ip.snapshot();
		int[] pixels = (int[])ip.getPixels();
		imp.trimProcessor();
		
		// convert to 8-bits
		long start = System.currentTimeMillis();
		MedianCut mc = new MedianCut(pixels, width, height);
		Image img8 = mc.convert(nColors);
	    imp.setImage(img8);
		IJ.showProgress(1.0);
	}
	
	/** Set true to scale to 0-255 when converting short to byte or float
		to byte and to 0-65535 when converting float to short. */
	public static void setDoScaling(boolean scaleConversions) {
		doScaling = scaleConversions;
	}

	/** Returns true if scaling is enabled. */
	public static boolean getDoScaling() {
		return doScaling;
	}
}
